#include "QUdev.h"
#include "QUdev_p.h"
#include <libudev.h>

#include <QDebug>

/**
  * @mainpage QUdev Library
  * @brief QUdev Library is a wrapper for the C udev library (libudev).
  * It is based on Qt4 and only requires the QtCore module and the
  * libudev itself.
  *
  * The API used the Qt4 style. The library was divided into the following
  * modules:
  * @list
  * -# QUdev - the main public class representing the udev subsystem.
  * -# QUdevDevice - representing a udev device.
  * -# QUdevEnumerator - a convenient class for retrieving (un)filtered list of devices.
  * -# QUdevMonitor - a convenient class for monitoring changes in the devices.
  * @endlist
  *
  * @footnote
  * @author Adrian Łubik
  *
  * @link
  * Website: http://sourceforge.net/projects/QUdev
  * @endlink
  * @endfootnote
  */


// Private Implemetation
// == QUdevContext:

inline static const char *cStr(const QString &string)
{
    return string.toLatin1().constData();
}

/**
  * @class QUdevContext
  * Encapsulates the pointer to a core udev struct.
  * @note This class is not accessible publicly and
  * is only used by the library classes. Only a const
  * reference to it can be obtained.
  */

/**
  * Creates a new instance of the class.
  * A new core udev struct will be created.
  */
QUdevContext::QUdevContext()
    : m_coreContext(udev_new())
{
//    qDebug() << "Constructed QUdevContext (new) ..." << this;
}

/**
  * Creates a copy of another instance.
  * Uses the pointer to the core udev struct from
  * the given copy. The core udev struct is referenced
  * at this point and dereferenced when this object is
  * destroyed.
  */
QUdevContext::QUdevContext(const QUdevContext &other)
    : m_coreContext(0)
{
    if (other.coreContext())
    {
        udev_ref(other.coreContext());
        m_coreContext = other.coreContext();
    }

//    qDebug() << "Constructed QUdevContext (copy) ..." << this;
}

/**
  * Destroys the object instance.
  * The core udev struct is dereferenced at this point.
  * If the core udev struct is not used by any other
  * copy of this context, it will automatically be released.
  */
QUdevContext::~QUdevContext()
{
    if (coreContext())
        udev_unref(coreContext());

//    qDebug() << "Destructed QUdevContext..." << this;
}

/**
  * Gives access to the core udev struct pointer.
  * @returns The pointer to the core udev struct.
  *
  * @warning If the pointer is intended to be used
  * outside the QUdevContext, it should be referenced
  * using udev_ref() and dereferenced by using udev_unref()
  * once it is not needed after external usage.
  */
struct udev *QUdevContext::coreContext() const
{
    return m_coreContext;
}


// Private Implementation
// == QUdevData:

QUdevData::QUdevData()
{
//    qDebug() << "Constructed QUDevPrivate..." << this;
}

QUdevData::QUdevData(const QUdevData &other)
    : QSharedData(other)
    , m_context(other.context())
{
//    qDebug() << "Constructed QUdevData from other..." << this;
}

QUdevData::~QUdevData()
{
//    qDebug() << "Destructed QUdevData..." << this;
}


struct udev *QUdevData::coreContext() const
{
    return context().coreContext();
}

const QUdevContext &QUdevData::context() const
{
    return m_context;
}

// Public Implementation
// == QUdev

/**
  * @class QUdev
  * @brief The main public class representing the udev subsystem.
  *
  * This is the main class representing the udev subsystem.
  * An instance of this class is needed to create objects which
  * depend on it (QUdevEnumerator, QUdevDevice, etc.).
  */

/**
  * Creates a new instance of the class.
  * Multiple instances are allowed. The data is
  * implicitly shared between copies of this class.
  *
  * @arg parent: The parent object.
  */
QUdev::QUdev(QObject *parent)
    : QObject(parent)
    , d(new QUdevData())
{
    if (d->coreContext() == NULL)
        qWarning("QUdev: Failed to create core context! This object is useless...");

//    qDebug() << "Constructed QUdev (new) ..." << this;
}

/**
  * Creates a copy of the given instance of QUdev.
  * The other.parent() will be used as the parent
  * for the new object.
  * It is very efficient to create copies as the
  * internal data is shared between them.
  */
QUdev::QUdev(const QUdev &other)
    : QObject(other.parent())
    , d(other.d)
{
//    qDebug() << "Constructed QUdev (copy)..." << this;
}


/**
  * Destroys the instance of the class.
  * If there are no more copies of this object
  * in use, the internal data will be released.
  */
QUdev::~QUdev()
{
//    qDebug() << "Destructing QUdev..." << this
            ;


//    qDebug() << "Destructed QUdev..." << this;
}

/**
  * The syspath of the current system configuration.
  *
  * @returns The path to the sys directory (usually /sys).
  */
QString QUdev::sysPath() const
{
    QString path = QLatin1String(udev_get_sys_path(d->coreContext()));
    return path;
}

/**
  * The devpath of the current system configuration.
  *
  * @returns The path to the dev directory (usually /dev).
  */
QString QUdev::devPath() const
{
    QString path = QLatin1String(udev_get_dev_path(d->coreContext()));
    return path;
}

/**
  * The runpath of the current system configuration.
  *
  * @returns The path to the run directory if the current version of
  * the udev library supports it. An empty string is returned otherwise.
  */
QString QUdev::runPath() const
{
#if UDEV_HAS_RUN_PATH
    QString path = QLatin1String(udev_get_run_path(d->coreContext()));
    return path;
#else
    return QString();
#endif
}

/**
  * The QUdevContext instance used by this object.
  * @returns A reference to the QUdevContext used by this instance.
  */
const QUdevContext &QUdev::context() const
{
    return d->context();
}

/**
  * A convenient function retrieving a device with the provided syspath.
  *
  * @arg syspath: The syspath representing the required device.
  * @arg parent: The parent object for the new QUdevDevice object.
  *
  * @see QUdevDevice::fromSyspath()
  * @returns A QUdevDevice object representing the device with the given syspath.
  */
QUdevDevice QUdev::deviceFromSysPath(const QString &syspath, QObject *parent) const
{
    struct udev *udev = d->coreContext();

    Q_ASSERT(udev != NULL);

    struct udev_device *coreDevice = udev_device_new_from_syspath(udev, cStr(syspath));
    QUdevDevice device(d->context(), coreDevice, parent);
    udev_device_unref(coreDevice); // QUdevDevice has called udev_device_ref() for its 'copy'

    return device;
}

/**
  * A convenient method retrieving a device based on the given subsystem.
  * Use this method if the desired subsystem is not specified in the QUdevNS::Subsystem enum,
  * i.e. when passing a custom subsystem.
  *
  * @arg subsystem: The name of the subsystem.
  * @arg sysname: The sysname of the device.
  * @arg parent: The parent object for the new QUdevDevice object.
  *
  * @see QUdevDevice::fromSubsystem()
  * @returns A QUdevDevice object representing the device with the given syspath.
  */
QUdevDevice QUdev::deviceFromSubsystem(const QString &subsystem, const QString &sysname, QObject *parent) const
{
    struct udev *udev = d->coreContext();

    Q_ASSERT(udev != NULL);

    struct udev_device *coreDevice = 0;
    coreDevice = udev_device_new_from_subsystem_sysname(udev, cStr(subsystem), cStr(sysname));

    QUdevDevice device(d->context(), coreDevice, parent);
    udev_device_unref(coreDevice); // QUdevDevice has called udev_device_ref() for its 'copy'

    return device;
}

/**
  * An overloaded convenient method retrieving a device based on the given subsystem.
  * This method takes an enum value for the subsystem. This one is recommended to use
  * as it eliminates a misspelling risk.
  *
  * @arg subsystem: The name of the subsystem.
  * @arg sysname: The sysname of the device.
  * @arg parent: The parent object for the new QUdevDevice object.
  *
  * @see QUdevDevice::fromSubsystem()
  * @returns A QUdevDevice object representing the device with the given syspath.
  */
QUdevDevice QUdev::deviceFromSubsystem(QUdevNS::Subsystem subsystem, const QString &sysname, QObject *parent) const
{
    return QUdev::deviceFromSubsystem(QUdevNS::subsystemToString(subsystem), sysname, parent);
}

/**
  * Retrieves the currently set logging priority.
  * @returns Current logging priority.
  */
int QUdev::loggingPriority() const
{
    struct udev *udev = d->coreContext();

    Q_ASSERT(udev != NULL);

    return udev_get_log_priority(udev);
}

/**
  * Sets the current logging priority. This value controls which
  * messages are logged.
  *
  * @arg priority: The new priority value.
  */
void QUdev::setLoggingPriority(int priority) const
{
    struct udev *udev = d->coreContext();

    Q_ASSERT(udev != NULL);

    udev_set_log_priority(udev, priority);
}

QUdev &QUdev::operator =(const QUdev &other)
{
    d = other.d;
    setParent(other.parent());

    return (*this);
}
