#include "QUdevDevice_p.h"
#include "QUdev.h"

#include <libudev.h>

#include <QDebug>

QUdevDeviceData::QUdevDeviceData(const QUdevContext &context, struct udev_device *device)
    : m_context(context)
    , m_device(0)
{
    setDevice(device);
}

QUdevDeviceData::QUdevDeviceData(const QUdevDeviceData &other)
    : QSharedData(other)
    , m_context(other.context())
    , m_device(0)
{
    setDevice(other.device());
}

QUdevDeviceData::~QUdevDeviceData()
{
    setDevice(NULL);
}

struct udev_device *QUdevDeviceData::device() const {
    return m_device;
}

void QUdevDeviceData::setDevice(struct udev_device *device) {
    if (this->device()) {
//        qDebug("setDevice called. unreferencing old core device.");
        udev_device_unref(this->device());
    }

    udev_device_ref(device);
    m_device = device;
}

const QUdevContext &QUdevDeviceData::context() const
{
    return m_context;
}

// Public Implementation:

/** @class QUdevDevice
  * @brief A class representing a udev device.
  *
  * A class representing a device retrieved from the udev subsystem.
  * A new device object can be obtained in a few ways:
  * @list
  * -# By calling one of the static methods of this class.
  * -# By calling one of the convenient method on a QUdev instance object.
  * -# By retrieving devices via the QUdevEnumerator.
  * @endlist
  *
  * In the first two cases a device syspath needs to be provided or
  * a subsystem and and a sysname of the device.
  *
  * If a requested device does not exist, the isValid() method will
  * return false.
  */

/**
  * Creates a new instance of this class using the given
  * context and the pointer to the core udev_device struct.
  * This constructor is used internally by QUdev Library.
  *
  * @arg context: A reference to the QUdevContext instance to be used by this device.
  * @arg device: A valid pointer to the core udev_device struct.
  * @arg parent: The parent object.
  *
  */
QUdevDevice::QUdevDevice(const QUdevContext &context, udev_device *device, QObject *parent)
    : QObject(parent)
    , d(new QUdevDeviceData(context, device))
{
//    qDebug() << this << "New QUdevDevice with new d_ptr:" << d_ptr;
}

/**
  * Creates a copy of an existing instance.
  * This is very efficient since the internal data is
  * implicitly shared by the copies.
  */
QUdevDevice::QUdevDevice(const QUdevDevice &other)
    : QObject(other.parent())
    , d(other.d)
{
//    qDebug() << this << "New QUdevDevice from copy of" << &other << "d_pointer referenced:" << d_ptr;
}

/**
  * Destructs the instance.
  * If there are no more copies of this object,
  * the internal data will be released.
  */
QUdevDevice::~QUdevDevice()
{
}

/**
  * Assigns another QUdevDevice to this object.
  * Since the internal data is implicitly shared, this
  * is very efficient. The old internal data is dereferenced
  * and if it is not used by any other copy, it will be released.
  */
QUdevDevice &QUdevDevice::operator=(const QUdevDevice &other)
{
    if (other.d == d)
    {
        qDebug() << "QUdevDevice tried to assign itself to itself!";
        return (*this);
    }

    d = other.d;
    return (*this);
}


/**
  * Retrieves a list of all properties of this device.
  *
  * @returns A list of all device properties from the udev subsystem.
  * An empty hash table will be returned if the device does not exist.
  */
QUdevProperties QUdevDevice::properties() const {
    QUdevProperties devProperties;

    if (!d->device())
        return devProperties;

    struct udev_list_entry *entry = udev_device_get_properties_list_entry(d->device());

    while (entry != NULL) {
        QLatin1String name(udev_list_entry_get_name(entry));
        QLatin1String value(udev_list_entry_get_value(entry));

        devProperties.insert(name, QVariant(value));
        entry = udev_list_entry_get_next(entry);
    }

    return devProperties;
}

/**
  * Checks if the device represented by this object exists in the udev subsystem.
  * @returns True if the device exists, false otherwise.
  */
bool QUdevDevice::isValid() const
{
    return (d->device() != NULL);
}

/**
  * Fetches the parent device of the device represented by
  * this object.
  *
  * If the subsystem and/or devtype have a non-empty value,
  * the parent device matching those will be looked-up.
  *
  * @arg subsystem: Optional. If given, only parent matching the given subsystem
  * will be looked-up.
  * @arg devtype: Optional. If given, onlu parent matching the given devtype
  * will be looked-up.  *
  *
  * @returns A QUdevDevice object representing the parent device and matching specified criteria (if any).
  * @note If there is no parent device at all or no parent device matching given criteria (if any),
  * the returned object's isValid() method will return false.
  */
QUdevDevice QUdevDevice::parentDevice(const QString &subsystem, const QString &devtype) const
{
    struct udev_device *coreDevice = d->device();

    if (subsystem.isEmpty() && devtype.isEmpty())
        return (QUdevDevice(d->context(), udev_device_get_parent(coreDevice), parent()));

    return (QUdevDevice(d->context(), udev_device_get_parent_with_subsystem_devtype(coreDevice, L1STR2(subsystem), L1STR2(devtype)), parent()));
}

/**
  * Retrieves the name of the driver assigned to this device.
  *
  * @returns The driver name if there is one assigned to the device.
  * An empty string will be returned if no drivers are assigned.
  */
QString QUdevDevice::driver() const
{
    return QLatin1String(udev_device_get_driver(d->device()));
}

/**
  * Creates a new device from syspath using the given QUdev class instance.
  *
  * @arg udev: A reference to the QUdev object to be used by this device.
  * @arg syspath: The syspath of the requested device.
  * @arg parent: The parent object for the new QUdevDevice instance.
  *
  * @returns The created device. If no such device was found, the isValid()
  * method of the created object will return false.
  */
QUdevDevice QUdevDevice::fromSyspath(const QUdev &udev, const QString &syspath, QObject *parent)
{
    return udev.deviceFromSysPath(syspath, parent);
}

/**
  * Creates a new device from syspath and sysname using the given QUdev class instance.
  *
  * @arg udev: A reference to the QUdev object to be used by this device.
  * @arg subsystem: The subsystem of the requested device.
  * @arg sysbame: The sysname of the requested device.
  * @arg parent: The parent object for the new QUdevDevice instance.
  *
  * @returns The created device. If no such device was found, the isValid()
  * method of the created object will return false.
  */
QUdevDevice QUdevDevice::fromSubsystem(const QUdev &udev, const QString &subsystem, const QString &sysname, QObject *parent)
{
    return udev.deviceFromSubsystem(subsystem, sysname, parent);
}

/**
  * Creates a new device from syspath and sysname using the given QUdev class instance.
  * This overloaded method takes subsystem as a QUdevNS::Subsystem enum value.
  *
  * @arg udev: A reference to the QUdev object to be used by this device.
  * @arg subsystem: The subsystem of the requested device.
  * @arg sysbame: The sysname of the requested device.
  * @arg parent: The parent object for the new QUdevDevice instance.
  *
  * @returns The created device. If no such device was found, the isValid()
  * method of the created object will return false.
  */
QUdevDevice QUdevDevice::fromSubsystem(const QUdev &udev, QUdevNS::Subsystem subsystem, const QString &sysname, QObject *parent)
{
    return udev.deviceFromSubsystem(subsystem, sysname, parent);
}

/**
  * @returns The devnode of the device.
  */
QString QUdevDevice::devnode() const
{
    if (!d->device())
        return QString();

    QString node = QLatin1String(udev_device_get_devnode(d->device()));
    return node;
}

/**
  * If the object was created due to a triggered action,
  * this method will return the name of that action.
  * An empty string will be returned otherwise.
  *
  * @returns The name of the action that this object was created by.
  */
QString QUdevDevice::action() const
{
    if (!d->device())
        return QString();

    QString action = QLatin1String(udev_device_get_action(d->device()));
    return action;
}

/**
  * @returns The devpath of this device.
  */
QString QUdevDevice::devpath() const
{
    if (!d->device())
        return QString();

    QString path = QLatin1String(udev_device_get_devpath(d->device()));
    return path;
}

/**
  * @returns The subsystem of this device.
  */
QString QUdevDevice::subsystem() const
{
    if (!d->device())
        return QString();

    QString system = QLatin1String(udev_device_get_subsystem(d->device()));
    return system;
}

/**
  * @returns The sysname of this device.
  */
QString QUdevDevice::sysname() const
{
    if (!d->device())
        return QString();

    QString name = QLatin1String(udev_device_get_sysname(d->device()));
    return name;
}

/**
  * @returns The sysnum of this device.
  */
QString QUdevDevice::sysnum() const
{
    if (!d->device())
        return QString();

    QString num = QLatin1String(udev_device_get_sysnum(d->device()));
    return num;
}

/**
  * @returns The syspath of this device.
  */
QString QUdevDevice::syspath() const
{
    if (!d->device())
        return QString();

    QString path = QLatin1String(udev_device_get_syspath(d->device()));
    return path;
}

/**
  * @returns The type of this device.
  */
QString QUdevDevice::type() const
{
    if (!d->device())
        return QString();

    QString t = QLatin1String(udev_device_get_devtype(d->device()));
    return t;
}

/**
  * Looks up the value for the given device property.
  *
  * @arg property: The name of the property.
  * @arg decode: Indicates whether to decode spaces or not of properties ending
  * with the _ENC suffix.
  *
  * @returns The value of the given property.
  */
QVariant QUdevDevice::value(const QString &property, bool decode) const
{
    if (!d->device())
        return QVariant();

    QString val = QLatin1String(udev_device_get_property_value(d->device(), L1STR(property)));

    if (decode && property.endsWith(QLatin1String("_ENC")) && val.contains("\\x20"))
        return QVariant(val.replace(QLatin1String("\\x20"), QLatin1String(" ")));

    return QVariant(val);
}

/**
  * Looks up the value for the given device property.
  * This overloaded method takes the property as a QUdevNS::Property enum value.
  *
  * @arg property: The name of the property.
  * @arg decode: Indicates whether to decode spaces or not of properties ending
  * with the _ENC suffix.
  *
  * @returns The value of the given property.
  */
QVariant QUdevDevice::value(QUdevNS::Property property, bool decode) const
{
    return this->value(propertyToString(property), decode);
}

/**
  * @returns The tags associated with this device.
  */
QStringList QUdevDevice::tagList() const
{
    QStringList list;

    if (!d->device())
        return list;

    QUdevNS::fillQStringList(list, udev_device_get_tags_list_entry(d->device()));
    return list;
}

/**
  * Lists sysattributes of this device.
  *
  * @warning This only works if the QUdev Library is
  * compiled against libudev 170 or newer. With older
  * versions it will always print a warning and return an empty list.
  *
  * @returns The list of sysattributes.
  */
QStringList QUdevDevice::sysattrList() const
{
#if UDEV_HAS_SYSATTR_LIST_ENTRY
    QStringList list;

    if (!d->device())
        return list;

    QUdevNS::fillQStringList(list, udev_device_get_sysattr_list_entry(d->device()));
    return list;
#else
    qWarning("QUdevDevice::sysattrList(): This version of QUdev was compiled against an old libudev not supporting this option!\n"
             "                            This option requires libudev 170 or newer.");
    return QStringList();
#endif
}

/**
  * Checks if the device has been initialized
  *
  * @warning This only works if the QUdev Library is
  * compiled against libudev 170 or newer. With older
  * versions it will always print a warning and return false.
  *
  * @returns True if the device was initialized.
  */
bool QUdevDevice::isInitialized() const
{
#if UDEV_HAS_IS_INITIALIZED
    if (!d->device())
        return false;

    return (1 == udev_device_get_is_initialized(d->device()));
#else
    qWarning("QUdevDevice::isInitialized(): This version of QUdev was compiled against an old libudev not supporting this option!\n"
             "                              This option requires libudev 170 or newer.");
    return false;
#endif
}

/**
  * @returns The seqnum of the device.
  */
seqnum_t QUdevDevice::seqnum() const
{
    if (!d->device())
        return 0;

    return udev_device_get_seqnum(d->device());
}

/**
  * The number of microseconds passed since udev set up the device for the first time.
  *
  * This is only implemented for devices with need to store properties in the udev database.
  * All other devices return 0 here.
  *
  * @warning This only works if the QUdev Library is
  * compiled against libudev 170 or newer. With older
  * versions it will print a warning and return 0.
  *
  * @returns The time elapsed since the device was initialized or 0 in described conditions.
  */
usec_t QUdevDevice::timeSinceInitialized() const
{
#if UDEV_HAS_USEC_SINCE_INITIALIZED
    if (!d->device())
        return 0;

    return udev_device_get_usec_since_initialized(d->device());
#else
    qWarning("QUdevDevice::timeSinceInitialized(): This version of QUdev was compiled against an old libudev not supporting this option!\n"
             "                                     This option requires libudev 170 or newer.");
    return 0;
#endif
}

void QUdevDevice::printInfo() const
{
    qDebug() << "Is valid:" << isValid();

    if (!isValid())
        return;

    qDebug() << "devnode:  " << devnode();
    qDebug() << "devpath:  " << devpath();
    qDebug() << "driver:   " << driver();
    qDebug() << "subsystem:" << subsystem();
    qDebug() << "sysname:  " << sysname();
    qDebug() << "sysnum:   " << sysnum();
    qDebug() << "syspath:  " << syspath();
    qDebug() << "type:     " << type();
    qDebug() << "tags:     " << tagList();
}
