#include "QUdevEnumerator.h"
#include "QUdevEnumerator_p.h"

#include "QUdev.h"
#include "QUdev_p.h"

#include <libudev.h>

#include <QDebug>

inline static void fillQUdevDeviceList(const QUdev &udev, QUdevDeviceList &list, struct udev_list_entry *firstEntry)
{
    struct udev_list_entry *entry = firstEntry;

    while (entry)
    {
        QString syspath = QLatin1String(udev_list_entry_get_name(entry));
        list.append(udev.deviceFromSysPath(syspath));

        entry = udev_list_entry_get_next(entry);
    }
}

// Private Implementation:
QUdevEnumeratorPrivate::QUdevEnumeratorPrivate(const QUdev &udev)
    : m_udev(udev)
    , m_enumerator(udev_enumerate_new(m_udev.context().coreContext()))
{
}

QUdevEnumeratorPrivate::~QUdevEnumeratorPrivate()
{
    if (m_enumerator)
        udev_enumerate_unref(m_enumerator);
}

const QUdev &QUdevEnumeratorPrivate::udev() const
{
    return m_udev;
}

const QUdevContext &QUdevEnumeratorPrivate::context() const
{
    return udev().context();
}

struct udev_enumerate *QUdevEnumeratorPrivate::coreEnumerator() const
{
    return m_enumerator;
}

// Public Implementation:

/**
  * @class QUdevEnumerator
  * @brief A convenient class for retrieving (un)filtered list of devices.
  *
  * QUdevEnumerator is a class used to obtain the list of devices
  * recognized by udev. It allows scanning all devices, as well as
  * subsystems. Different filters can be applied to return only
  * those devices the programmer is interested in.
  */

/**
  * @enum QUdevEnumerator::FilterOption
  *  - Match
  *      - Filters devices matching the criteria.
  *  - DoNotMatch
  *      - Filters devices which do not match the given filter.
  * @endlist
  */

/**
  * @enum QUdevEnumerator::ScanOption
  * @list
  * - ScanAllDevices
  *     - Indicates scanning of all devices in the udev subsystem.
  * - ScanSubsystems
  *     - Indicates scanning of devices in subsystems in the udev subsystem.
  */

/**
  * @var typedef QUdevDeviceList QHash<QString,QString>
  * @brief QHash<QString, QString>
  */

/**
  * Creates a new instance of the class. This class does not
  * allow copies of itself.
  * @arg udev: A reference to a QUdev class instance.
  * @arg parent: Parent object.
  */
QUdevEnumerator::QUdevEnumerator(const QUdev &udev, QObject *parent)
    : QObject(parent)
    , d_ptr(new QUdevEnumeratorPrivate(udev))
{
    if (d_ptr->coreEnumerator() == NULL)
        qWarning("QUdevEnumerator: cannot obtain core enumerator - this object is useless!");
}

/**
  * Destroys the object. All devices retrieved from this object
  * will still be valid.
  */
QUdevEnumerator::~QUdevEnumerator()
{
    delete d_ptr;
}

/**
  * Applies a subsystem filter to the list of looked up devices.
  * This filter can be applied multiple number of times. In such case
  * devices (not)matching either of the given subsystems will be returned.
  *
  *
  * @arg subsystem: The filter to be applied in looking up devices.
  *
  * @arg filterOption: Specifies whether to look up devices matching
  * the given filter, or the ones not matching the filter. By default
  * the QUdevEnumerator::Match option is applied.
  *
  *
  * @returns True if the filter was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addFilterSubsystem(const QString &subsystem, QUdevEnumerator::FilterOption filterOption)
{
    if (subsystem.isEmpty())
    {
        qDebug("addFilterSubsystem: Cannot add an empty filter!");
        return false;
    }

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    if (filterOption == QUdevEnumerator::Match)
        return (udev_enumerate_add_match_subsystem(d->coreEnumerator(), L1STR(subsystem)) == 0);
    else
        return (udev_enumerate_add_nomatch_subsystem(d->coreEnumerator(), L1STR(subsystem)) == 0);
}

bool QUdevEnumerator::addFilterSubsystem(QUdevNS::Subsystem filter, FilterOption filterOption)
{
    return this->addFilterSubsystem(QUdevNS::subsystemToString(filter), filterOption);
}

/**
  * Applies a sysattribute filter to the list of looked up devices.
  * This filter can be applied multiple number of times. In such case
  * devices (not)matching either of the given sysattributes will be returned.
  *
  *
  * @arg attribute: The filter to be applied in looking up devices.
  *
  * @arg value: The optional value for the attribute. A QString("*")
  * should be passed if only the attribute name is required - in this
  * case any value will be matched. Simple wildcards work as well.
  *
  * Applying an empty string (QString("")) will look up for values
  * consisting of an empty string, NOT those not having any value.
  *
  * @arg filterOption: Specifies whether to look up devices matching
  * the given filter, or the ones not matching the filter. By default
  * the QUdevEnumerator::Match option is applied.
  *
  *
  * @returns True is returned if the filter was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addFilterSysattr(const QString &attribute, const QString &value, QUdevEnumerator::FilterOption filterOption)
{
    if (attribute.isEmpty())
    {
        qDebug("addFilterSysattr: Cannot add an empty filter!");
        return false;
    }

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    if (filterOption == QUdevEnumerator::Match)
        return (udev_enumerate_add_match_sysattr(d->coreEnumerator(), L1STR(attribute), L1STR(value)) == 0);
    else
        return (udev_enumerate_add_nomatch_sysattr(d->coreEnumerator(), L1STR(attribute), L1STR(value)) == 0);
}


/**
  * Applies a property filter to the list of looked up devices.
  * This filter can be applied multiple number of times. In such case
  * devices matching either of the given properties will be returned.
  *
  *
  * @arg property: The filter to be applied in looking up devices.
  *
  * @arg value: The optional value for the attribute. A QString("*")
  * should be passed if only the attribute name is required - in this
  * case any value will be matched. Simple wildcards work as well.
  *
  * Applying an empty string (QString("")) will look up for values
  * consisting of an empty string, NOT those not having any value.
  *
  *
  * @returns True is returned if the filter was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addFilterProperty(const QString &property, const QString &value)
{
    if (property.isEmpty())
    {
        qDebug("addFilterProperty: Cannot add an empty filter!");
        return false;
    }

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    return (udev_enumerate_add_match_property(d->coreEnumerator(), L1STR(property), L1STR(value)) == 0);
}

bool QUdevEnumerator::addFilterProperty(QUdevNS::Property property, const QString &value)
{
    return this->addFilterProperty(propertyToString(property), value);
}

/**
  * Applies a tag filter to the list of looked up devices.
  * This filter can be applied multiple number of times. In such case
  * devices matching either of the given tags will be returned.
  *
  *
  * @arg filter: The filter to be applied in looking up devices.
  *
  * @returns True is returned if the filter was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addFilterTag(const QString &filter)

{
    if (filter.isEmpty())
    {
        qDebug("addFilterTag: Cannot add an empty filter!");
        return false;
    }

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    return (udev_enumerate_add_match_tag(d->coreEnumerator(), L1STR(filter)) == 0);
}

/**
  * Applies a sysname filter to the list of looked up devices.
  * This filter can be applied multiple number of times. In such case
  * devices matching either of the given sysnames will be returned.
  *
  *
  * @arg sysname: The filter to be applied in looking up devices.
  *
  * @returns True is returned if the filter was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addFilterSysname(const QString &sysname)
{
    if (sysname.isEmpty())
        return false;

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    return (udev_enumerate_add_match_sysname(d->coreEnumerator(), L1STR(sysname)) == 0);
}


/**
  * Adds a syspath of the device that will be included in the
  * list of looked up devices, provided that a device with the
  * given syspath exists.
  *
  * @arg syspath: The syspath of the device to be included in lookup
  * results.
  *
  * @returns True is returned if the syspath was applied successfully, false
  * otherwise.
  */
bool QUdevEnumerator::addSyspath(const QString &syspath)
{
    if (syspath.isEmpty())
    {
        qDebug("addSyspath: Cannot add an empty syspath!");
        return false;
    }

    Q_D(const QUdevEnumerator);

    if (d->coreEnumerator() == NULL)
        return false;

    return (udev_enumerate_add_syspath(d->coreEnumerator(), L1STR(syspath)) == 0);
}

/**
  * Resets the enumerator to its original state.
  * All filters are removed and if these are required
  * they should be applied again. This is usefull to
  * invoke before executing a new lookup based on
  * different criteria to the previous ones.
  *
  * @returns True if the object was reset successfully.
  *
  * @warning If this method returns false, that means
  * this object is not useful anymore and no further lookups
  * will succeed.
  */
bool QUdevEnumerator::reset()
{
    if (!d_ptr)
        return false;

    QUdev udev = d_ptr->udev();
    delete d_ptr;
    d_ptr = new QUdevEnumeratorPrivate(udev);

    return (d_ptr->coreEnumerator() != NULL);
}

/**
  * Lists syspaths of looked-up devices. Depending on filters
  * applied, this will be a list of syspaths of all found devices.
  *
  * @arg scanOption: indicates what type of device scan will be called.
  * this can be either an all devices or a subsystems scan.
  * @see QUdevEnumerator::ScanOption.
  *
  * @returns A QStringList with syspaths of found devices.
  */
QStringList QUdevEnumerator::listDevicePaths(QUdevEnumerator::ScanOption scanOption) const
{
    Q_D(const QUdevEnumerator);
    QStringList list;

    if (d->coreEnumerator() == NULL)
        return list;

    if (scanOption == ScanAllDevices)
    {
        if (udev_enumerate_scan_devices(d->coreEnumerator()) != 0)
            return list;
    }
    else
    {
        if (udev_enumerate_scan_subsystems(d->coreEnumerator()) != 0)
            return list;
    }

    struct udev_list_entry *firstEntry = udev_enumerate_get_list_entry(d->coreEnumerator());
    QUdevNS::fillQStringList(list, firstEntry);

    return list;
}

/**
  * Lists looked-up devices. Depending on filters
  * applied, this will be a list of all found devices.
  *
  * @arg scanOption: indicates what type of device scan will be called.
  * this can be either an all devices or a subsystems scan.
  * @see QUdevEnumerator::ScanOption.
  *
  * @returns A QUdevDeviceList with objects representing found devices.
  */
QUdevDeviceList QUdevEnumerator::listDevices(QUdevEnumerator::ScanOption scanOption) const
{
    Q_D(const QUdevEnumerator);
    QUdevDeviceList list;

    if (d->coreEnumerator() == NULL)
        return list;

    if (scanOption == ScanAllDevices)
    {
        if (udev_enumerate_scan_devices(d->coreEnumerator()) != 0)
            return list;
    }
    else
    {
        if (udev_enumerate_scan_subsystems(d->coreEnumerator()) != 0)
            return list;
    }

    struct udev_list_entry *firstEntry = udev_enumerate_get_list_entry(d->coreEnumerator());
    fillQUdevDeviceList(d->udev(), list, firstEntry);

    return list;
}
