#include "QUdevMonitor.h"
#include "QUdevMonitor_p.h"
#include "QUdev.h"

#include <QDebug>
#include <QSocketNotifier>

#include <libudev.h>

static const char *__QUDEV_SRC_KERNEL = "kernel";
static const char *__QUDEV_SRC_UDEV   = "udev";


inline static const char *eventSourceToCStr(const QUdevMonitor::EventSource &s)
{
    if (s == QUdevMonitor::Udev)
        return __QUDEV_SRC_UDEV;
    else
        return __QUDEV_SRC_KERNEL;
}

// Private Implementation:

QUdevMonitorPrivate::QUdevMonitorPrivate(QUdevMonitor *parent, QUdevMonitor::EventSource eventSource, const QUdevContext &context)
    : q_ptr(parent)
    , m_context(context)
    , m_eventSource(eventSource)
    , m_notifier(0)
    , m_coreMonitor(0)
    , m_isListening(false)
{
    Q_ASSERT(q_ptr != NULL);
    reset();
}

QUdevMonitorPrivate::~QUdevMonitorPrivate()
{
    if (m_notifier)
        delete m_notifier;

    if (coreMonitor())
        udev_monitor_unref(coreMonitor());
}

struct udev_monitor *QUdevMonitorPrivate::coreMonitor() const
{
    return m_coreMonitor;
}

bool QUdevMonitorPrivate::reset()
{
    if (m_notifier)
    {
        delete m_notifier;
        m_notifier = 0;
    }

    if (coreMonitor())
        udev_monitor_unref(coreMonitor());

    m_coreMonitor = udev_monitor_new_from_netlink(m_context.coreContext(), eventSourceToCStr(m_eventSource));
    m_isListening = false;

    if (!coreMonitor())
    {
        qWarning("QUdevMonitorPrivate: Cannot create new core monitor - this object is useless!");
        return false;
    }

    Q_Q(QUdevMonitor);
    Q_ASSERT(q != NULL);

    m_notifier = new QSocketNotifier(udev_monitor_get_fd(coreMonitor()), QSocketNotifier::Read);
    m_notifier->connect(m_notifier, SIGNAL(activated(int)), q, SLOT(processEvent()));

    return true;
}

const QUdevContext &QUdevMonitorPrivate::context() const
{
    return m_context;
}


// Public Implementation:

QUdevMonitor::QUdevMonitor(const QUdev &udev, QUdevMonitor::EventSource eventSource, QObject *parent)
    : QObject(parent)
    , d_ptr(new QUdevMonitorPrivate(this, eventSource, udev.context()))
{
}

QUdevMonitor::~QUdevMonitor()
{
    if (d_ptr)
        delete d_ptr;
}

void QUdevMonitor::processEvent()
{
    Q_D(QUdevMonitor);

    if (d->coreMonitor() == NULL)
        return;

    struct udev_device *coreDevice = udev_monitor_receive_device(d->coreMonitor());
    QUdevDevice device(d->context(), coreDevice);
    udev_device_unref(coreDevice);

    if (!device.isValid())
        return;

    QString action = device.action();
    qDebug() << "Received new action:" << action << "for device" << device.syspath();

    if (action == QLatin1String("add"))
    {
        emit deviceAdded(device);
        emit deviceAdded(device.syspath());
        emit deviceChanged(device, static_cast<int>(QUdevMonitor::DeviceAdded));
    }
    else if (action == QLatin1String("remove"))
    {
        emit deviceRemoved(device);
        emit deviceRemoved(device.syspath());
        emit deviceChanged(device, static_cast<int>(QUdevMonitor::DeviceRemoved));
    }
    else if (action == QLatin1String("online"))
    {
        emit deviceOnline(device);
        emit deviceOnline(device.syspath());
        emit deviceChanged(device, static_cast<int>(QUdevMonitor::DeviceOnline));
    }
    else if (action == QLatin1String("change"))
    {
        emit deviceChanged(device);
        emit deviceChanged(device.syspath());
        emit deviceChanged(device, static_cast<int>(QUdevMonitor::DeviceChanged));
    }
    else
    {
        emit deviceChanged(device, static_cast<int>(QUdevMonitor::DeviceOtherAction));
    }
}

bool QUdevMonitor::isListening() const
{
    Q_D(const QUdevMonitor);
    return d->m_isListening;
}

bool QUdevMonitor::startListening()
{
    Q_D(QUdevMonitor);
    Q_ASSERT(d->m_notifier != NULL);

    if (d->m_isListening || !d->coreMonitor())
        return false;

    if (udev_monitor_enable_receiving(d->coreMonitor()) == 0)
        d->m_isListening = true;

    d->m_notifier->setEnabled(true);

    return isListening();
}

bool QUdevMonitor::stopListening()
{
    if (isListening())
    {
        Q_D(QUdevMonitor);
        Q_ASSERT(d->m_notifier != NULL);

        d->m_notifier->setEnabled(false);
        d->m_isListening = false;
    }

    return !isListening();
}

bool QUdevMonitor::reset()
{
    Q_D(QUdevMonitor);
    return d->reset();
}

bool QUdevMonitor::addFilterSubsystem(const QString &subsystem, const QString &devtype) const
{
    Q_D(const QUdevMonitor);

    if (isListening())
    {
        qWarning("Tried to apply a filter on currently listening monitor!");
        return false;
    }

    if (d->coreMonitor() == NULL)
        return false;

    return (0 == udev_monitor_filter_add_match_subsystem_devtype(d->coreMonitor(), L1STR(subsystem), L1STR(devtype)));
}

bool QUdevMonitor::addFilterSubsystem(QUdevNS::Subsystem subsystem, const QString &devtype) const
{
    return this->addFilterSubsystem(QUdevNS::subsystemToString(subsystem), devtype);
}

bool QUdevMonitor::clearFilters() const
{
    Q_D(const QUdevMonitor);

    if (d->coreMonitor() == NULL)
        return false;

    if (0 != udev_monitor_filter_remove(d->coreMonitor()))
        return false;

    return (0 == udev_monitor_filter_update(d->coreMonitor()));
}
